{                                                                           }
{ File:       Slack.dpr                                                     }
{ Function:   Program to calculate slack caused by cluster size.            }
{ Language:   Delphi for Win32                                              }
{ Author:     Rudy Velthuis                                                 }
{ Copyright:  (c) 2005 Rudy Velthuis                                        }
{ Disclaimer: This code is freeware. All rights are reserved.               }
{             This code is provided as is, expressly without a warranty     }
{             of any kind. You use it at your own risk.                     }
{                                                                           }
{             If you use this code, please credit me.                       }
{                                                                           }

program Slack;

{$APPTYPE CONSOLE}

uses SysUtils, Windows;

var
  ClusterSize: Word;
  StartDirectory: string;
  FileCount: Int64;
  DirectorySize: Int64;
  FileSystemName: array[0..9] of Char;

function GetClusterSize(Drive: Char): Longint;
var
  SectorsPerCluster,
  BytesPerSector,
  Dummy: Cardinal;
begin
  SectorsPerCluster := 0;
  BytesPerSector := 0;
  GetDiskFreeSpace(PChar(Drive + ':\'), SectorsPerCluster,
    BytesPerSector, Dummy, Dummy);
  Result := SectorsPerCluster * BytesPerSector;
  GetVolumeInformation(PChar(Drive + ':\'), nil, 0, nil, Dummy, Dummy,
    FileSystemName, SizeOf(FileSystemName));
end;

{ Calculates the wasted bytes for a file with the given size. }
function CalcWaste(Size: Longint): Longint;
var
  Diff: Longint;
begin
  Diff := Size mod ClusterSize;
  if Diff > 0 then
    CalcWaste := ClusterSize - Diff
  else
    CalcWaste := 0;
end;

{ Calculates the wasted bytes of all files in a directory and }
{ its subdirectories.                                         }
function CalcDirWaste(Dir: string): Int64;
var
  SR: TSearchRec;
  Error: Integer;
begin
  Result := 0;
  Dir := IncludeTrailingPathDelimiter(Dir);
  Error := SysUtils.FindFirst(Dir + '*.*', faAnyFile, SR);
  try
    while Error = 0 do
    begin
      if (SR.Attr and faDirectory) = 0 then
      begin
        Inc(Result, CalcWaste(SR.Size));
        Inc(FileCount);
        Inc(DirectorySize, SR.Size);
      end
      else if (SR.Name <> '.') and (SR.Name <> '..') then
        Inc(Result, CalcDirWaste(Dir + SR.Name));
      Error := SysUtils.FindNext(SR);
    end;
  finally
    SysUtils.FindClose(SR);
  end;
end;

procedure Help;
begin
  Writeln;
  Writeln('Usage: SLACK [Options] Filename');
  Writeln;
  Writeln('Options start with - or /');
  Writeln('  -H or -?  Show this help');
  Writeln;
  Writeln('SLACK calculates the number of wasted bytes caused by cluster size.');
  Writeln;
  Halt;
end;

function FormatInt(L: Int64): string;
begin
  Result := Format('%.n', [L + 0.0]);
end;

var
  Wasted: Int64;
  ParamNo: Integer;
  Param: string;

begin
  if ParamCount = 0 then
    GetDir(0, StartDirectory)
  else
  begin
    for ParamNo := 1 to ParamCount do
    begin
      Param := ParamStr(ParamNo);
      if (Param[1] in ['-', '/']) and (Length(Param) > 1) then
        case UpCase(Param[2]) of
          'H', '?': Help;
          else
            Help;
        end
      else
        StartDirectory := ExpandFileName(ParamStr(ParamNo));
    end;
  end;
  if not DirectoryExists(StartDirectory) then
  begin
    Writeln;
    Writeln('Invalid directory: ', StartDirectory);
    Halt(1);
  end;
  DirectorySize := 0;
  FileCount := 0;
  ClusterSize := GetClusterSize(StartDirectory[1]);
  Writeln;
  Writeln('Path:              ', StartDirectory);
  Writeln('File system type:  ', FileSystemName);
  Writeln('Cluster size:      ', FormatInt(ClusterSize), ' bytes');
  Write('Calculating wasted disk space, please wait...');
  Wasted := CalcDirWaste(StartDirectory);
  Write(#13'                                             '#13);
  Writeln('Total file sizes:  ', FormatInt(DirectorySize),
          ' bytes = ', FormatInt(DirectorySize div 1024), ' kb');
  Writeln('Wasted:            ', FormatInt(Wasted),
          ' bytes = ', FormatInt(Wasted div 1024), ' kb');
  Writeln('Percentage wasted: ', (100.0 * Wasted / (DirectorySize + Wasted)):0:1,
          '%');
  Writeln('Files:             ', FormatInt(FileCount), ', wasted ', FormatInt(Wasted div FileCount), ' bytes per file' );
end.

